/*
 *	scponly.c
 *
 * 	http://sublimation.org/scponly
 *	joe@sublimation.org
 *
 *	version 2.0: jul 2nd, 2002
 */

#include <stdio.h>	// io
#include <string.h>	// for str*
#include <unistd.h>	// for exit
#include <stdlib.h>	// atexit
#include <errno.h>
#include "scponly.h"

FILE *log=NULL;
int debuglevel=0;
int winscp_mode=0;
int chrooted=0;
char username[MAX_USERNAME];
char homedir[FILENAME_MAX];

int process_ssh_request(char *request);
int winscp_regular_request(char *request);
int winscp_transit_request(char *request);
int process_winscp_requests(void);

void cleanup (void)
{
	if (log != NULL)
		fclose(log);
}

int main (int argc, char **argv) 
{
	FILE *debugfile;
	
	/*
	 * set debuglevel.  any nonzero number will result in debugging info to log
	 */
	if (NULL!=(debugfile=fopen(DEBUGFILE,"r")))
	{
		fscanf(debugfile,"%u",&debuglevel);
		fclose(debugfile);
	}
	if (NULL==(log=fopen(LOGFILE,"a+")))
	{
		log=stderr;
		if (debuglevel) 
		{
			fprintf (log,"couldn't open logfile for writing: \"%s\"\n",LOGFILE);
			perror("fopen");
		}
		exit(-1);
	}
	/*
	 *	is this a chroot'ed scponly installation?
	 */
	if (0==strncmp(argv[0],CHROOTED_NAME,FILENAME_MAX))
	{
		if (debuglevel)
			fprintf(log,"chrooted binary in place, will chroot()\n");
		fflush(log);
		chrooted=1;
	}

	if (debuglevel)
	{
		int i;
		fprintf (log, "%d arguments in total.\n",argc);
		for (i=0;i<argc;i++)
			fprintf (log, "\targ %u is %s\n",i,argv[i]);
		fflush(log);
	}
	if (debuglevel)
		fprintf (log,"opened logfile %s\n",LOGFILE);
	fflush(log);
	atexit(cleanup);
	if (getuid()==0)
	{	
		log_stamp();
		fprintf(log,"root login denied\n");
		exit(-1);
	}
#ifdef WINSCP_COMPAT
	if ((argc!=3) && (argc!=1))
#else
	if (argc!=3)
#endif
	{
		if (debuglevel)
			fprintf (log,"incorrect number of args\n");
		exit(-1);
	}
	if (!get_uservar())
	{
		fprintf (log, "%s is misconfigured. contact sysadmin.\n",argv[0]);
		exit (-1);
	}
	if (chrooted)
	{
		if (debuglevel)
			fprintf (log,"chrooting to dir: \"%s\"\n",homedir);
		if (-1==(chroot(homedir)))
		{
			if (debuglevel)
			{
				fprintf (log,"couldn't chroot errno=%d\n",errno);
				perror("chroot");
				fflush(log);
			}
			log_stamp();
			fprintf(log,"couldn't chroot to %s\n",homedir);
			exit(-1);
		}
	}
	if (debuglevel)
		fprintf (log,"setting uid to %u\n",getuid());
	fflush(log);
	if (-1==(seteuid(getuid())))
	{
		fprintf (log,"couldn't revert to my real uid\n");
		perror("seteuid");
		exit(-1);
	}
#ifdef WINSCP_COMPAT
	if (argc==1)
	{
		if (debuglevel)
		{
			log_stamp();
			fprintf (log,"entering WinSCP compatibility mode\n");
			printf ("entering WinSCP compatibility mode\n");
			fflush(log);
		}
		if (-1==process_winscp_requests())
		{
			log_stamp();
			fprintf (log,"failed WinSCP compatibility mode\n");
			fflush(log);
			exit(-1);
		}
	}
#else
	if (0)	{}	// placeholder
#endif
	else if (-1==process_ssh_request(argv[2]))
	{
		log_stamp();
		fprintf (log,"bad request: %s\n",argv[2]);
		fflush(log);
		exit(-1);
	}
	if (debuglevel)
		fprintf (log,"scponly completed\n");
	fflush(log);
	exit(0);
}

#ifdef WINSCP_COMPAT

int winscp_transit_request(char *request)
{
	char *new_request=NULL;

	/*
	 * for scp -t and scp -f commands, winscp preceeds each
	 * command with a "begin-of-file" marker we need to
	 * provide, everything after that can be handled by
	 * winscp_regular_request.
	 */

	new_request=strbeg(request,WINSCP_BOF_REQ);
	if (NULL == new_request)
	{
		return(-1);	 // improper transit cmd
	}
	printf ("%s\n",WINSCP_BOF); // start transfer
	fflush(stdout);
	return(winscp_regular_request(new_request));
}

int winscp_regular_request(char *request)
{
	/*
	 * winscp uses one of two requests to terminate each command.
	 * we must determine which (if any) is terminating this request. fun!
	 */
	char *new_request=NULL;
	int retval=0;
	int retzero=1;
	new_request=strend(request, WINSCP_EOF_REQ_ZERO);
	if (NULL == new_request)
	{
		retzero=0;
		new_request=strend(request, WINSCP_EOF_REQ_RETVAL);
		if (NULL == new_request)
		{
			printf ("command wasn't terminated with %s or %s\n",
				WINSCP_EOF_REQ_RETVAL, WINSCP_EOF_REQ_ZERO);
			return(-1);	// bogus termination
		}
	}
	/*
	 *  ignore unalias and unset commands
	 */
//	fprintf (stderr, "new_request: %s\n",new_request);
	if ((NULL!=strbeg(new_request,"unset ")) ||
		(NULL!=strbeg(new_request,"unalias ")))
	{
		retval=0;
	}
	else
	{
		retval=process_ssh_request(new_request);
		if (retzero)		// ignore actual retval if winscp wants us to
			retval=0;
	}
	free(new_request);
	return(retval);
}

int process_winscp_requests(void)
{
        char    linebuf[MAX_REQUEST];
	int	count=0;		// num of semicolons in cmd
	int	ack=0;

	winscp_mode=1;

	//printf ("%s0\n",WINSCP_EOF); // reply to initial echo request
	fflush(stdout);

	/*
	 *	now process commands interactively
 	 */
        while (fgets(&linebuf[0],sizeof(linebuf), stdin) != NULL)
	{
		ack=0;

		if (strlen(linebuf)==0)
			return(-1);

		linebuf[strlen(linebuf)-1]=0;	// drop the trailing CR
		count=cntchr(linebuf,';');

		if (count==1)	// regular cmd
		{
			ack=winscp_regular_request(linebuf);
		}
		else if (count==2)	 // transit command
		{
			ack=winscp_transit_request(linebuf);
		}
		else
			ack=0; 	// winscp always sends 2 or 3 cmds at once

		printf ("%s%d\n",WINSCP_EOF,ack); // respond to client
		fflush(stdout);
	}
	return 0;
}

#endif

int exec_request(char *request)
{
	char tempbuf[MAX_REQUEST];
	int err;

	/*
	 * the following little nugget prevents
	 * attempts to overflow the cmd buf
	 */
	strncpy(tempbuf,request,MAX_REQUEST);
	tempbuf[MAX_REQUEST-1]='\0';

	/*
	 * I realize system is a lame version of
	 * execve, but I actually dont want to 
	 * handle wildcard expansion or argument
	 * parsing. leave that to /bin/sh
	 */
	if (err=system(tempbuf))
	{
		if (debuglevel)
			perror("system");	
		log_stamp();
		fprintf(log,"system() fail: %s, %s errno=%d\n",request,strerror(errno),errno);
	}
	return(err);
}

int process_ssh_request(char *request)
{
	if (debuglevel)
	{
		fprintf (log,"processing request: \"%s\"\n", request);
	}

	if (!valid_chars(request))
		return(-1);

	/*
	 *	some scp clients like to request a specific path for
	 *	sftp-server.  i will strip that out here and rely
	 *	upon sh's $PATH to find the binary 
	 */
	request=clean_request(request);

	/*
	 *	request must be one of a few commands.
	 *	some commands must be exact matches.
	 *	others only require the beginning of the string to match.
	 */

	if (NULL!=strbeg(request,"cd "))
	{
#ifdef WINSCP_COMPAT
		int retval;
		char *destdir=NULL;
		int reqlen=strlen(request);
		destdir=(char *)malloc(reqlen);
		if (destdir == NULL)
		{
			perror("malloc");
			exit(-1);
		}
		
		/*
		 * well, now that scponly is a persistent shell
		 * i have to maintain a $PWD.  damn.
		 * we're going to INSIST upon a double quote
		 * encapsulated new directory to change to.
		 */
		if (request[(reqlen-1)]=='"')
		{
			if (request[3]=='"') // check first doublequote
			{	
				bzero(destdir,reqlen);
				strncpy(destdir,&request[4],reqlen-5);
				retval=chdir(destdir);
				free(destdir);
				return(retval);
			}
		}


#endif
	}
	else if ((exact_match(request,"ls")) || 
		(exact_match(SCP2_ARG,request)) || 
#ifdef WINSCP_COMPAT
		(exact_match(request,"pwd")) ||
		(exact_match(request,"groups")) ||
		(NULL!=strbeg(request,"echo ")) ||
#endif
		(NULL!=strbeg(request,"ls ")) || 
		(NULL!=strbeg(request,"scp ")) ||
		(NULL!=strbeg(request,"rm ")) ||
		(NULL!=strbeg(request,"ln ")) ||
		(NULL!=strbeg(request,"mv ")) ||
		(NULL!=strbeg(request,"chmod ")) ||
		(NULL!=strbeg(request,"chown ")) ||
		(NULL!=strbeg(request,"mkdir ")) ||
		(NULL!=strbeg(request,"rmdir "))) 
			return(exec_request(request));

	/*
	 *	reaching this point in the code means the request isnt one of
	 *	our accepted commands
 	 */
	if (debuglevel)
	{
		fprintf (log,"denied request. not a supported command.\n");
		fprintf (log,"supported comands are: ls, scp, %s, pwd, chmod, mkdir, rm, mv, rmdir\n",SCP2_ARG);
	}
	log_stamp();
	fprintf (log,"denied request\n");

#ifdef WINSCP_COMPAT
	if (winscp_mode)
	{
		printf ("command not permitted by scponly\n");
		return(-1);
	}
	else
#endif 
	{
		exit(-1);
	}
}
